package edu.ntua.ivml.qa.lp;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import edu.ntua.ivml.qa.utils.Counter;
import edu.ntua.ivml.qa.utils.Utils;


public class Clause implements ClauseContainer, Cloneable, Comparable<Clause> {

	protected Atom head;
	protected ArrayList<Atom> body;
	
	private final String string;	
	
	private Map<Predicate, ArrayList<Integer>> predicateIndex;
	
	public Clause(Atom headx, Atom bodyx) {
		this.head = headx;
		
		this.body = new ArrayList<Atom>();
		body.add(bodyx);
		
		string = computeString();
	}
	
	public Clause(Atom headx, Collection<Atom> bodyx) {
		this.head = headx;

		this.body = new ArrayList<Atom>();
		for (Atom a : bodyx) {
			body.add(a);
		}
		
		string = computeString();
	}

	public String getString() {
		return string;
	}

	public Substitution replaceUnboundVariablesByUnderscoreSubstitution() {
		Substitution subst = new Substitution();
	
		Set<Variable> uVars = getUnboundVariables(); 
		
		for (Atom atom : body) {
			Term[] terms = atom.getArguments().getTerms();
			for (int i = 0; i < terms.length; i++) {
				Term t = terms[i];
				if (uVars.contains(t)) {
					if (i == 0) {
						subst.add((Variable)t, Variable.UNBOUND_VARIABLE_1);
					} else {
						subst.add((Variable)t, Variable.UNBOUND_VARIABLE_2);
					}
				}
			}
		}
		
		return subst;
	}

	public Substitution prefixUnboundVariablesByUnderscoreSubstitution() {
		Substitution subst = new Substitution();
	
		for (Variable v : getUnboundVariables()) {
			if (!v.hasUnboundPrefix()) {
				subst.add(v, new Variable(Variable.UNBOUND_PREFIX + v.getName()));
			}
		}
		
		return subst;
	} 
	
	public Clause prefixUnboundVariablesByUnderscore() {
		return apply(prefixUnboundVariablesByUnderscoreSubstitution());
	}	
	
	public Clause unprefixUnboundVariablesByUnderscore() {
		Substitution subst = new Substitution();
	
		for (Variable v : getVariableSet()) {
			if (v.hasUnboundPrefix()) {
				subst.add(v, new Variable(v.getName().substring(1)));
			}
		}
		
		return apply(subst);
	}	
	
    private String computeString(){
    	Atom[] sBody = body.toArray(new Atom[] {});
    	
    	Arrays.sort(sBody, new Comparator<Atom>(){
    		public int compare(Atom a1, Atom a2){
    			return a1.getStringPrefix().compareTo(a2.getStringPrefix());
    		}
        });

    	String res = head.toString();
        
        if(sBody.length > 0){
            res += " <- ";
             for (int i = 0; i < sBody.length; i++) {
            	 if (i != 0) {
            		res += ", ";
            	 }
            	 res += sBody[i].toString();
             }
        }
        
        return res;
    }
    
	public Object clone() {
		try {
			Clause c = (Clause)super.clone();
			c.head = (Atom)head.clone();
			c.body = new ArrayList<Atom>();
			for (int i = 0; i < body.size(); i++) {
				c.body.add((Atom)body.get(i).clone());
			}
			
			return c;
		} catch (CloneNotSupportedException e) {
			throw new InternalError();
		}				
	}		
	
	private Map<Predicate, ArrayList<Integer>> getPredicateIndex() {
		if (predicateIndex == null) {
			predicateIndex = computePredicateIndex();
		}
		
		return predicateIndex;
	}
	
	private HashMap<Predicate, ArrayList<Integer>> computePredicateIndex() {
		
		HashMap<Predicate, ArrayList<Integer>> index = new HashMap<Predicate, ArrayList<Integer>>();
		
		for (int i = 0; i < body.size(); i++){
			Predicate p = body.get(i).getPredicate();
		
			ArrayList<Integer> list = index.get(p);
			if (list == null) {
				list = new ArrayList<Integer>();
				index.put(p,list);
			}
			list.add(i);
		}

		return index;
	}
	
	public Clause getClause() {
		return this;
	}
	
	public Atom getHead() {
		return head;
	}
	
	public int bodySize() {
		return body.size();
	}

	public ArrayList<Atom> getBody() {
		return body;
	}
	
	public Atom getBodyAtomAt(int i) {
		return body.get(i);
	}

	public boolean isSubsumedBy(Clause r) {
		return r.subsumes(this);
	}

	private int bodyDepth = -1;
	
	public int bodyDepth() {
		if (bodyDepth == -1) {
			int maxDepth = 0;
		
			for (int i = 0; i < body.size(); i++){
				maxDepth = Math.max(maxDepth, body.get(i).depth());
			}
			
			bodyDepth = maxDepth;
		}
		
		return bodyDepth;
	}
	
    public boolean isFunctionFree() {
    	if (head.depth() == 0 && bodyDepth() == 0){
    		return true;
    	}
    	
    	return false;
    }
    
    public int[] termIndexPattern(Variable t) {
    	return AtomArray.termIndexPattern(body, t);
    }
   
    private Set<Variable> vars = null;
    private Set<Term> terms = null;

    private Set<Variable> getVariableSet() {
		if (vars == null) {
			Set<Variable> res = new HashSet<Variable>();
			res.addAll(head.getVariables());
			
			for (Atom b : body) {
				res.addAll(b.getVariables());
			}
			
			vars = res;
		}
		
		return vars;
    }
    
	public Set<Variable> getVariables() {
		
		getVariableSet();
		
		return (Set<Variable>)((HashSet<Variable>)vars).clone();
	}
	
	public Set<Term> getTerms() {
		if (terms == null) {
			Set<Term> res = new HashSet<Term>();
			res.addAll(head.getTerms());
			
			for (Atom b : body) {
				res.addAll(b.getTerms());
			}
			
			terms = res;
		} 
		
		return (Set<Term>)((HashSet<Term>)terms).clone();
	}	
	
	public Set<Predicate> getPredicatesInBody() {
		Set<Predicate> res = new HashSet<Predicate>();
		
		for (Atom a : body) {
			res.add(a.getPredicate());
		}
		
		return res;
	}
	
	
	public Clause apply(Substitution subst) {
		
		ArrayList<Atom> newBody = new ArrayList<Atom>();

		for (int i = 0; i < body.size(); i++) {
			Atom newAtom = body.get(i).apply(subst);
			if (!newBody.contains(newAtom)) {
				newBody.add(newAtom);
			}
		}
		
		
		return new Clause(head.apply(subst), newBody);
	}
	


	public int compareTo(Clause r) {
		int c = head.compareTo(r.getHead());
		if (c != 0) {
			return c;
		} else {
			ArrayList<Atom> cBody = r.getBody();
			for (int i = 0; i < Math.min(cBody.size(), body.size()); i++) {
				int nc = body.get(i).compareTo(cBody.get(i));
				if (nc != 0) {
					return nc;
				}
			}
			
			if (body.size() < cBody.size()) {
				return -1;
			} else if (body.size() > cBody.size()) {
				return 1;
			} else {
				return 0;
			}
		}
	}	
	
	public boolean equals(Object obj) {
		return isEquivalentTo((Clause)obj); 
	}

	public int hashCode() {
		int h = head.getPredicate().hashCode();

		for (int i = 0; i < bodySize(); i++) {
			h += getBodyAtomAt(i).getPredicate().hashCode();
		}

		return h;
	}
	
	private class ClauseSubstitution {
		Clause clause;
		Substitution subst;
		
		private ClauseSubstitution(Clause clause, Substitution subst) {
			this.clause = clause;
			this.subst = subst;
		}
		
		public int hashCode() {
			return clause.hashCode();
		}
		
		public boolean equals(Object obj) {
			if (obj instanceof ClauseSubstitution) {
				return clause.equals(((ClauseSubstitution)obj).clause);
			} else {
				return false;
			}
		}
	}
	
	public Clause condense() {

		ArrayList<Clause> clauses = new ArrayList<Clause>();
		ArrayList<Clause> condensed = new ArrayList<Clause>();
	
		clauses.add(this);
		
		while(!clauses.isEmpty()){
			Clause givenClause = clauses.remove(0);
			condensed.add(givenClause);

			int givenClauseSize = givenClause.bodySize();
			
			for(int i = 0; i < givenClauseSize - 1; i++){
				Atom givenClauseAtomI = givenClause.getBodyAtomAt(i);
				loop:
				for(int j = i+1; j < givenClauseSize; j++){
					Substitution unifier = Substitution.mgu(givenClauseAtomI, givenClause.getBodyAtomAt(j));
					if (unifier != null){
						Clause newQuery = givenClause.apply(unifier);
						
						for (Clause c: clauses){
							if (c.isEquivalentTo(newQuery)){
								continue loop;
							}					
						}
						
						for (Clause c: condensed){
							if (c.isEquivalentTo(newQuery)){
								continue loop;
							}
						}
						
						clauses.add(newQuery);
	                }
				}
			}
		}
		
		Collections.sort(condensed, new Comparator<Clause>(){
			public int compare(Clause c1, Clause c2){
			    return (new Integer(c1.bodySize())).compareTo(new Integer(c2.bodySize()));
			}
		});

		for(Clause c: condensed){
			if (c.subsumes(this)){
				return c;
			}
		}

		return this;
	
	}

	public Clause condenseWithOriginalVariables() {

		ArrayList<ClauseSubstitution> clauses = new ArrayList<ClauseSubstitution>();
		ArrayList<ClauseSubstitution> condensed = new ArrayList<ClauseSubstitution>();
	
		clauses.add(new ClauseSubstitution(this, new Substitution()));
		
		while(!clauses.isEmpty()){
			ClauseSubstitution givenClause = clauses.remove(0);
			condensed.add(givenClause);
			
			int givenClauseSize = givenClause.clause.bodySize();

			for(int i = 0; i < givenClauseSize - 1; i++){
				Atom givenClauseAtomI = givenClause.clause.getBodyAtomAt(i);
				loop:
				for(int j = i+1; j < givenClauseSize; j++){
					Substitution unifier = Substitution.mgu(givenClauseAtomI, givenClause.clause.getBodyAtomAt(j));
					if (unifier != null){
						Clause newQuery = givenClause.clause.apply(unifier);
						
						for (ClauseSubstitution c: clauses){
							if (c.clause.isEquivalentTo(newQuery)){
								continue loop;
							}					
						}
						
						for (ClauseSubstitution c: condensed){
							if (c.clause.isEquivalentTo(newQuery)){
								continue loop;
							}
						}
						
						clauses.add(new ClauseSubstitution(newQuery, givenClause.subst.compose(unifier)));
	                }
				}
			}
		}
		
		Collections.sort(condensed, new Comparator<ClauseSubstitution>(){
			public int compare(ClauseSubstitution c1, ClauseSubstitution c2){
			    return (new Integer(c1.clause.bodySize())).compareTo(new Integer(c2.clause.bodySize()));
			}
		});

		Set<Variable> unboundVars = this.getUnboundVariables();
		
		for(ClauseSubstitution c: condensed){
			if(c.clause.subsumes(this)){
				
				Substitution iSubst = new Substitution();
				
			    for (Variable v : c.subst.getVariables()) {
			    	if (unboundVars.contains(c.subst.getSubstitutionTerm(v)) && !unboundVars.contains(v)) {
			    		iSubst.add((Variable)c.subst.getSubstitutionTerm(v), v);
			    	}
			    }
			    
			    removeRedundantAtomsSubstitution = c.subst.compose(iSubst);

				return c.clause.apply(iSubst);
			}
		}
		
		return this;
	
	}
	
	public Substitution removeRedundantAtomsSubstitution;

	public Set<Variable> getBoundVariables() {
		if (bVars == null) {
			Set<Variable> res = getVariables();
			res.removeAll(getUnboundVariables());
			bVars = res;
		}
		
		return (Set<Variable>)((HashSet<Variable>)bVars).clone();
	}
	
	private Set<Variable> bVars = null;
	private Set<Variable> uVars = null;
	
	public Set<Variable> getUnboundVariables() {
		if (uVars == null) {
			Map<Variable, Counter> cc = new HashMap<Variable, Counter>();
			
			for (Variable v : head.getVariables()) {
				Counter c = cc.get(v);
				if (c == null) {
					c = new Counter(2);
					cc.put(v, c);
				} else {
					c.increase();
				}
			}
			
			Set<Atom> bodyAtoms = new HashSet<Atom>();
			
			for (Atom atom : body) {
				if (!bodyAtoms.add(atom)) {
					continue;
				}
				
				for (Term t : atom.getArguments().getTerms()) {
					if (t.isVariable()) {
						Counter c = cc.get((Variable)t);
						if (c == null) {
							c = new Counter(1);
							cc.put((Variable)t, c);
						} else {
							c.increase();
						}
					}
				}			
			}
			
			Set<Variable> res = new HashSet<Variable>();
			
			for (Map.Entry<Variable, Counter> entry : cc.entrySet()) {
				if (entry.getValue().getValue() == 1) {
					res.add(entry.getKey());
				}
			}
			
			uVars = res;
		}
		
		return (Set<Variable>)((HashSet<Variable>)uVars).clone();

	}	

	public <U extends ClauseContainer> boolean isCCSubsumedBy(U u) {
		return isSubsumedBy(u.getClause());
	}
	
	public static <U extends ClauseContainer> ArrayList<U> getNonRedundantClauses(ArrayList<U> clauses, boolean oneway, ArrayList<Integer> ids, ArrayList<Integer> blocks, ArrayList<Integer> blockIds) {
		return getNonRedundantClauses(clauses, oneway, null, ids, blocks, blockIds);
	}
	
	public static <U extends ClauseContainer> ArrayList<U> getNonRedundantClauses(ArrayList<U> clauses, boolean oneway, Counter ct, ArrayList<Integer> ids, ArrayList<Integer> block, ArrayList<Integer> blockIds) {
		ArrayList<U> result = new ArrayList<U>();
		result.addAll(clauses);

		ArrayList<Integer> newIds = null;
		ArrayList<Integer> blocks = null;
		ArrayList<Integer> blocksIds = null;
		
		if (ids != null) {
			newIds = new ArrayList<Integer>();
			newIds.addAll(ids);
			
			blocks = new ArrayList<Integer>();
			blocksIds = new ArrayList<Integer>();
			
			blocks.addAll(block);
			blocksIds.addAll(blockIds);
		}
		
		ClauseContainer cc;
		int id = -1;
		int bi = -1;

		int prevId = -1;
		
		if (!oneway) {
			loop1:
			for (int i = 0; i < result.size(); ) {
				cc = result.get(i);
				if (newIds != null) {
					id = newIds.get(i);

					if (blocks != null) {
						if (id != prevId) {
							bi = Utils.binarySearch(blocksIds, id);
							prevId = id;
						}
					}
				}
				
				for (int j = 0; j < i;) {
					if (ids != null && id == newIds.get(j)) {
						if (blocks != null) {
							if (bi + 1 < blocks.size()) {
								j = blocks.get(bi + 1);
							} else {
								break;
							}
						} else {
							j++;
						}

						continue;
					}
						
					if (cc.isCCSubsumedBy(result.get(j))) {
						if (ct != null && i < ct.getValue()) {
							ct.decrease();
						}
						
						result.remove(i);
						
						if (newIds != null) {
							newIds.remove(i);
							
							if (blocks != null) {
								for (int t = bi + 1; t < blocks.size(); t++) {
									blocks.set(t, blocks.get(t) - 1);
								}

								if (bi + 1 < blocks.size() && blocks.get(bi).equals(blocks.get(bi + 1))) {
									blocks.remove(bi);
									blocksIds.remove(bi);
								}
							}
						}
						
						continue loop1;
					}
					
					j++;
				}
				
				i++;
			}
		}
		
		prevId = -1;
		
		loop2:
		for (int i = 0; i < result.size();) {
			cc = result.get(i);
			if (newIds != null) {
				id = newIds.get(i);
				if (blocks != null) {
					if (id != prevId) {
						bi = Utils.binarySearch(blocksIds, id);
						prevId = id;
					}
				}
			}
			
			for (int j = i + 1; j < result.size();) {
				if (ids != null && id == newIds.get(j)) {
					if (blocks != null) {
						if (bi + 1 < blocks.size()) {
							j = blocks.get(bi + 1);
						} else {
							break;
						}
					} else {
						j++;
					}
					continue;
				}

				if (cc.isCCSubsumedBy(result.get(j))) {
					if (ct != null && i < ct.getValue()) {
						ct.decrease();
					}
					result.remove(i);

					if (newIds != null) {
						newIds.remove(i);
						
						if (blocks != null) {
							for (int t = bi + 1; t < blocks.size(); t++) {
								blocks.set(t, blocks.get(t) - 1);
							}

							if (bi + 1 < blocks.size() && blocks.get(bi).equals(blocks.get(bi + 1))) {
								blocks.remove(bi);
								blocksIds.remove(bi);
							}
						}					
					}
					
					continue loop2;
				}
				
				j++;
			}
		
			i++;
		}
		
		return result;
	}	
	
	public boolean isQueryClause() {
		return getHead().getPredicate().isQueryPredicate();
	}
	
	public String toString() {
		String s = "";
		
		for (Atom b : body) {
			if (s.length() != 0) {
				s += ", ";
			}
			s += b.toString();
		}
		
		return head.toString() + " <- " + s;
	}	
	
	
    private boolean canSubsume(Clause that) {

		if (!head.getPredicate().equals(that.getHead().getPredicate())) {
			return false;
		}
		
		Set<Predicate> thatPreds = new HashSet<Predicate>();
		int thatBodySize = that.bodySize();
		
		for (int i = 0; i < thatBodySize; i++){			
			thatPreds.add(that.getBodyAtomAt(i).getPredicate());
		}
		
		for (Predicate p : this.getPredicateIndex().keySet()) {
			if (!thatPreds.contains(p)) {
				return false;
			}
		}			
		return true;
    }	
    
    private boolean canbeEquivalent(Clause that){
    	
		if (!head.getPredicate().equals(that.getHead().getPredicate()) || body.size() != that.bodySize()) {
			return false;
		}

		Set<Predicate> allPreds = new HashSet<Predicate>();
		for (int i = 0; i < this.body.size(); i++){
			allPreds.add(this.body.get(i).getPredicate());
			allPreds.add(that.body.get(i).getPredicate());
		}
		
		for (Predicate p : allPreds) {
			ArrayList<Integer> l1 = this.getPredicateIndex().get(p);			
			ArrayList<Integer> l2 = that.getPredicateIndex().get(p);
			
			if (l1 == null || l2 == null || l1.size() != l2.size()) {
				return false;
			}
		}
		
		return true;
    }	
	
    
    public boolean subsumes(Clause rule) {
    	if (!canSubsume(rule)) {
    		return false;
    	}
    	
        return match(false, new Atom[][] { new Atom[]{head}, body.toArray(new Atom[] {}) },
        		     createAtomIndex(rule.getHead()), createAtomIndex(rule.getBody()), 
        		     new Substitution(), 0, 0);
    }
    
    public boolean isEquivalentTo(Clause rule) {
    	if (!canbeEquivalent(rule)){
    		return false;
    	}
    	
        return match(true, new Atom[][] { new Atom[]{head}, body.toArray(new Atom[] {}) },
        		     createAtomIndex(rule.getHead()), createAtomIndex(rule.getBody()),
        		     new Substitution(), 0, 0);
    }	
	
    private boolean match(boolean subs, Atom[][] atoms, Map<Predicate,IndexNode> headIndex, Map<Predicate,IndexNode> bodyIndex, Substitution substitution, int headBodyIndex, int index) {
        Atom[] currentAtoms = atoms[headBodyIndex];
        
        if (index == currentAtoms.length) {
            if (headBodyIndex == 0) {
                return match(subs, atoms, headIndex, bodyIndex, substitution, 1, 0);
            } else {
                return true;
            }
        } else {
            Map<Predicate, IndexNode> currentIndex;
            
            if (headBodyIndex == 0) {
            	currentIndex = headIndex;
            } else {
            	currentIndex = bodyIndex;
            }

            Atom atom = currentAtoms[index];
            IndexNode nodes = currentIndex.get(atom.getPredicate());
            
            while (nodes != null) {
            	if (subs){
                    if (!nodes.matched) {
                        Substitution matchedSubstitution = matchAtoms(atom, nodes.atom, substitution);
                        if (matchedSubstitution != null) {
                            nodes.matched = true;
                            if (match(subs, atoms, headIndex, bodyIndex, matchedSubstitution, headBodyIndex, index + 1)) {
                                return true;
                            }
                            nodes.matched = false;
                        }
                    }
                    nodes = nodes.next;
            	} else {
            		Substitution matchedSubstitution = matchAtoms(atom, nodes.atom, substitution);
                    if (matchedSubstitution != null) {
                        if (match(subs, atoms, headIndex, bodyIndex, matchedSubstitution, headBodyIndex, index + 1)) {
                            return true;
                        }
                    }
                    nodes = nodes.next;
            	}
            }
            return false;
        }
    }

    private static Map<Predicate, IndexNode> createAtomIndex(Atom atom) {
    	ArrayList<Atom> atoms = new ArrayList<Atom>();
    	atoms.add(atom);
    	
        return createAtomIndex(atoms);
    }	

    private static Map<Predicate, IndexNode> createAtomIndex(ArrayList<Atom> atoms) {
        Map<Predicate, IndexNode> atomIndex = new HashMap<Predicate, IndexNode>();
        
        for (Atom atom : atoms) {
        	atomIndex.put(atom.getPredicate(), new IndexNode(atom, atomIndex.get(atom.getPredicate())));
        }
        
        return atomIndex;
    }	

    private Substitution matchAtoms(Atom a1, Atom a2, Substitution subst) {
        
    	boolean substitutionCopied = false;
        
        for (int i = 0; i < a1.getPredicate().getArity(); i++) {
            Term t1 = a1.getArguments().getTerm(i);
            Term t2 = a2.getArguments().getTerm(i);
            
            if (t1.isVariable()) {
                Term existingBinding = subst.getSubstitutionTerm((Variable)t1);
                
                if (existingBinding == null) {
                    if (!substitutionCopied) {
                        subst = (Substitution)subst.clone();
                        substitutionCopied = true;
                    }
                    subst.add((Variable)t1, t2);
                }
            }
            
            if (!t1.apply(subst).equals(t2)) {
                return null;
            }
        }
        
        return subst;
    }
    
    private static class IndexNode {
        private final Atom atom;
        private final IndexNode next;
        private boolean matched;

        public IndexNode(Atom atom, IndexNode next) {
            this.atom = atom;
            this.next = next;
            this.matched = false;
        }
    }

}


