package edu.ntua.ivml.qa.lp;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;


public class Substitution implements Cloneable {
	
	private HashMap<Variable, Term> subst;
	
	public Substitution() {
		subst = new HashMap<Variable, Term>();
	}
	
	public void clear() {
		subst.clear();
	}

	public boolean isEmpty() {
		return subst.isEmpty();
	}
	
	public Object clone() {
		try {
			Substitution s = (Substitution)super.clone();
			s.subst = (HashMap<Variable, Term>)subst.clone();
			
			return s;
		} catch (CloneNotSupportedException e) {
			throw new InternalError();
		}				
	}	
	
	public boolean add(Substitution newSubst) {
		
		for (Variable v : newSubst.getVariables()) {

			Term tOrig = this.getSubstitutionTerm(v);			
			Term tNew = newSubst.getSubstitutionTerm(v);
			
			if (tOrig != null && tNew != null && !tOrig.equals(tNew)) {
				return false;
			}
			
			if (tOrig == null && tNew != null) {
				this.add(v, tNew);		
			}
		}
		
		return true;
		
	}
			
	public static Substitution createDisjointRenaming(Set<Variable> keep, Set<Variable> change) {
		Substitution subst = new Substitution();
		
		for (Variable v : change) {

			if (keep.contains(v)) {
				int i = 1;
				Variable newVar = new Variable(v.getName() + "_" + i);
				while (keep.contains(newVar) || change.contains(newVar)) {
					newVar = new Variable(v.getName() + "_" + (++i));
				}
				subst.add(v, newVar);
			}
		}
		
		return subst;
	}
	
	public void remove(Variable var) {
		subst.remove(var);
	}
	
	public boolean add(Variable var, Term t) {
		Term v = (Term)subst.get(var);
			
		if (v != null) {
			if (!v.equals(t)) {
				return false;
			} else {
				return true;
			}
		} else {
			subst.put(var, t);	
			return true;
		}
	}
		
	public Term getSubstitutionTerm(Variable var) {
		return (Term)subst.get(var);
	}
	
	public Set<Variable> getVariables() {
		return subst.keySet();
	}
	
	private static ArrayList<Term> disagreementSet(Tuple... tuples) {
		ArrayList<Term> res = new ArrayList<Term>();
		
		int s = tuples[0].size();

		for (int j = 0; j < s; j++) {
			Term t = tuples[0].getTerm(j);
			
			for (int i = 1; i < tuples.length; i++) {
				if (!tuples[i].getTerm(j).equals(t)) {
					res.add(tuples[i].getTerm(j).firstDifferenceTerm(t));
					for (int k = 1; k < tuples.length; k++) {
						Term d = t.firstDifferenceTerm(tuples[k].getTerm(j));
						if (d != null && !res.contains(d)) {
							res.add(d);
						}
					}
					
					return res;
				}
			}
		}
		
		return res;
			
	}
	
	public Substitution compose(Substitution s2) {
		Substitution res = new Substitution();
		
		for (Map.Entry<Variable, Term> s : subst.entrySet()) {
			Variable v = s.getKey();
			Term t = s.getValue().apply(s2);
			if (!t.equals(v)) {
				res.add(v, t);
			}
		}

		Set<Variable> vars = getVariables();
		
		for (Map.Entry<Variable, Term> s : s2.subst.entrySet()) {
			Variable v = s.getKey();
			Term t = s.getValue();
			if (!vars.contains(v)) {
				res.add(v, t);
			}
		}
		
		return res;
		
	}
	
	public Set<Variable> getInverseVariables(Term t, Set<Variable> sourceVariables) {
		Set<Variable> res = new HashSet<Variable>();
		
		for (Map.Entry<Variable, Term> entry : subst.entrySet()) {
			if (t.equals(entry.getValue()) && sourceVariables.contains(entry.getKey())) {
				res.add(entry.getKey());
			}
		}
		
		return res;
	}
	
	public static Substitution mgu(Atom... atoms) {
		
		Predicate p = atoms[0].getPredicate();
		
		Tuple[] tuples = new Tuple[atoms.length];
		
		tuples[0] = atoms[0].getArguments();
		
		for (int i = 1 ; i < atoms.length; i++) {
			if (!atoms[i].getPredicate().equals(p)) {
				return null;
			} else {
				tuples[i] = atoms[i].getArguments();
			}
		}
		
		return mgu(tuples);
	
	}
	
	public Substitution restrict(Set<Variable> vars) {
		Substitution res = new Substitution();
		
		for (Map.Entry<Variable, Term> entry : subst.entrySet()) {
			if (vars.contains(entry.getKey())) {
				res.add(entry.getKey(), entry.getValue());
			}
		}
		
		return res;
		
	}

	public void removeFunctionalBindings() {
		for (Iterator<Map.Entry<Variable, Term>> iter = subst.entrySet().iterator(); iter.hasNext();) {
			Map.Entry<Variable, Term> entry = iter.next();
			if (entry.getValue().isFunctionalTerm()) {
				iter.remove();
			}
		}
	}
	
	private Map<Variable, Term> getSubstitutionMap() {
		return subst;
	}
	
	public boolean containsFunctions() {
		for (Map.Entry<Variable, Term> entry : subst.entrySet()) {
			if (entry.getValue().depth() > 0) {
				return true;
			}
		}

		return false;
	}
	
	public boolean isRenaming(Set<Variable> tvars) {

		Substitution rs = restrict(tvars);
		
		Set<Variable> ys = new HashSet<Variable>();
		
		for (Term term : rs.getSubstitutionMap().values()) {
			if (!term.isVariable() || ys.contains(term)) {
				return false;
			}
			 
			ys.add((Variable)term);
		}
		
		Set<Variable> xs = rs.getSubstitutionMap().keySet();
		Set<Variable> es = new HashSet<Variable>();
		es.addAll(tvars);

		if (!es.containsAll(xs)) {
			return false;
		}
		
		es.removeAll(xs);
		
		int s = ys.size();
		ys.removeAll(es);
		
		if (ys.size() != s) {
			return false;
		}
		
		return true;
		
	}
	
	public static Substitution mgu(Tuple... tuples) {

		for (int i = 0; i < tuples.length; i++) {
			if (tuples[i].size() != tuples[0].size()) {
				return null;
			}
		}
		
		Substitution subst = new Substitution();

		Tuple[] sTuples = new Tuple[tuples.length];
		
		for (;;) {
			
			for (int i = 0; i < tuples.length; i++) {
				sTuples[i] = tuples[i].apply(subst);
			}
			
			ArrayList<Term> ds = disagreementSet(sTuples);
		
			if (ds.isEmpty()) {
				return subst;
			}
			
			boolean unify = false;

			loop:
			for (int i = ds.size() - 1; i >= 0; i--) {
				Term t1 = ds.get(i);
				
				for (int j = ds.size() - 1; j >=0 ; j--) {
					if (i == j) {
						continue;
					}
					
					Term t2 = ds.get(j);
					
					if (t1.isVariable() && !t2.getVariables().contains(t1)) {
						Substitution s2 = new Substitution();
						s2.add((Variable)t1, t2);
						subst = subst.compose(s2);
						unify = true;
						break loop;
					}
				}
			}
			
			if (!unify) {
				return null;
			}
			
		}
	}
	
	public String toString() {
		String s = "";
		
		for (Map.Entry<Variable, Term> entry : subst.entrySet()) {
			if (s.length() != 0) {
				s += ", ";
			}
			s += entry.getKey().toString() + "/" + entry.getValue().toString();
		}
		
		return "{" + s + "}";
	}

}